//************************************************************************************
//* Mega2560_OV7670_test.c
//*
//* C-code source file containing the main program
//*
//* Created: 17.09.2014 16:56:55
//* Last modification: 11.12.2014
//*
//* Author: Dirk Uffmann (uffmann@arcor.de)
//* 
//* This program should run on an Arduino Mega 2560 R3 board with an Lseeduino TFT 
//* shield V1.3 and 2.4 inch TFT V2.1 with 320x240 pixel.
//* Intention is to test an OV7670 Camera module by Alientek V2.2 with FIFO AL422B.
//* For debugging purpose a text LCD with I2C bus adapter is connected via TWI.
//*
//* The hardware and software is documented in an article provided in the internet
//* magazine ft:pedia 04/2014 by www.ftcommunity.de in german language
//*
//* Title of the article: "Einfache Ziffernerkennung mit einer CMOS-Kamera am AVR-
//* Controller am Beispiel eines ft-Kreditkartenleser-Modells" 
//*
//* This software is provided as is without any warranty, guarantee or liability.
//* The author takes no responsibility for consequences or damage caused by the usage
//* of this software. 
//*
//************************************************************************************ 

#include <avr/io.h>
#include <stdlib.h>
#include <stdio.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <math.h>
#include "i2c2master.h" //second i2c bus master for 3,3 V devices (OV7670)
#include "i2clcd.h" // text LCD
#include "OV7670.h"
#include "ILI9325D.h"
#include "image_processing.h"

const char line2_right[] PROGMEM="Zahl";
const char line1_left[] PROGMEM="t_x_l"; 
const char line2[] PROGMEM="t_x_r";
const char line3[] PROGMEM="t_y_t";
const char line4[] PROGMEM="t_y_b";

ISR (TIMER0_OVF_vect)
{
	// timing is adapted for system clock 16 MHz with prescaler divide by 8 -> 8 * 256 / (16 MHz) = 128 s or 7.9 kHz
	// the variable timer_IR is exclusively used in INT2 ISR for IR control.
	// do not change timer_IR in other functions
	timer0 += 1; //defined in i2clcd.h
}

void init (void)
{
/* description of IO-pins:
 */	
	DDRA = 0; // Camera input
	PORTA = 0; // no pullups
		
	DDRD = _BV(PD3) + _BV(PD7);	// Pin 3 (FIFO_WEN) enables writing to the FIFO (active high because of NAND gate on Alientek Camera Module)
								// Pin 7 (TFT Register Select RS) 
	PORTD = _BV(PD7);			// Pin 7 high, Pin3 initially set low (FIFO write enable inactive)

	DDRF = _BV(PF2) + _BV(PF3) + _BV(PF4);	// Pin4 (RCLK), Pin 3 (OE) and 2 (RRST) on port F are used for read clock, output enable and read reset of FIFO (the two latter are low active)
	PORTF = _BV(PF2) + _BV(PF3) + _BV(PF4);	// outputs are initially set high (both are low active)

	DDRG = _BV(PG0) + _BV(PG1) + _BV(PG2);	// Pin 0 on Port G is used as TFT Reset
											// Pin 1 on Port G is used as chip select for TFT	
											// Pin 2 on port G is used for read clock of AL422 FIFO and also as write clock for the color TFT LCD
	PORTG = _BV(PG0) + _BV(PG1) + _BV(PG2);	// Reset for TFT set high (i.e. iactive)
									// CS for TFT set high (not selected) 
									// TFT write strobe WS initially set high

//	system clock 16 MHz is used (external quartz oscillator)

//	start timer0 with prescaler clock/8
	TCCR0B = _BV(CS01);

//	activate 8-bit Timer0 overflow interrupt 16 MHz / 256 / 8 = 7812 Hz = 128 s
	TIMSK0 = _BV(TOIE0);
	

//	idle mode is set up as sleep mode
	SMCR = _BV(SE);

//	rising edge on INT2 generates an interrupt
	EICRA = _BV(ISC21) +_BV(ISC20);
    sei ();
}


int main(void)
{
	uint16_t figures = 0;
	
	init(); // initialize ATMEGA2560R3 Arduino board

	FIFO_read_reset();

	i2c_init(); // initialize I2C serial communication for LCD text display
	lcd_init(); // initialize LCD text display

	TFT_init();
	CAM_init();

	lcd_command(LCD_CLEAR);                                 //-     Clear Screen
	lcd_wait_ms(3);											//-		the clear screen command takes 1,5 ms
	lcd_gotolr(1,1);
	lcd_print_p(line1_left);
	lcd_gotolr(2,1);
	lcd_print_p(line2);
	lcd_gotolr(2,15);
	lcd_print_p(line2_right);
	lcd_gotolr(3,1);
	lcd_print_p(line3);
	lcd_gotolr(4,1);
	lcd_print_p(line4);
	
// endless loop for image capture, image display and image processing 
	do
	{
		figures = identify_figures(40);
//		"40" is the threshold for filtering the image
//		adapt this threshold according to your illumination
		
		LC_display_dec(figures, 3, 15);

	} while (1);

	i2c_stop();
	
	while(1) sleep_mode();
}
