//*************************************************************************************
//* C-code source file "OV7670.c" with functions for Alientek V2.2 camera module using 
//* Omnivision OV7670 and FIFO AL422 
//*
//* Created: 17.09.2014 16:56:55
//* Last modification: 11.12.2014
//*
//* Author: Dirk Uffmann (uffmann@arcor.de)
//*
//* This program should run on an Arduino Mega 2560 R3 board with an Lseeduino TFT
//* shield V1.3 and 2.4 inch TFT V2.1 with 320x240 pixel.
//* Intention is to test an OV7670 Camera module by Alientek V2.2 with FIFO AL422B.
//* For debugging purpose a text LCD with I2C bus adapter is connected via TWI.
//*
//* The hardware and software is documented in an article provided in the internet
//* magazine ft:pedia 04/2014 by www.ftcommunity.de in german language
//*
//* Title of the article: "Einfache Ziffernerkennung mit einer CMOS-Kamera am AVR-
//* Controller am Beispiel eines ft-Kreditkartenleser-Modells"
//*
//* This software is provided as is without any warranty, guarantee or liability.
//* The author takes no responsibility for consequences or damage caused by the usage
//* of this software.
//*
//*************************************************************************************

#include "OV7670.h"

const unsigned char cam_init_data[] PROGMEM = {
	REG_CLKRC, 0x81, // clock prescaler (divide by 2), i.e. f_int = 48MHz/2 = 24 MHz
	REG_DBLV, 0x4A, // PLLx4 i.e. 12 MHz Quartz -> 48 MHz

	REG_COM3, 0x04, // enable down-sampling
	REG_COM7, 0x04, // select RGB output
	REG_RGB444, 0x00, // please, no RGB444
	REG_COM15, 0xD0, // select RGB565 output format and full output range
	REG_COM10, 0x02,	// access register 0x15 = COM10
	//set VSYNC to negative

	//	This should give an image 160x120 (VGA scaled by factor 4 in both dimensions)
	//	to store this image we would need 2 bytes x 160 x 120 i.e. 38.400 bytes

	REG_COM14, 0x1A, //enable PCLK scaling and down-sampling and divide PCLK by 8
	0x72, 0x22, //down-sample by 8
	0x73, 0xF2, //divide by 8
	0x3A,0x00, //TSLB set output window manually

	//	No windowing, full field of view
	REG_HSTART,0x16,
	REG_HSTOP,0x04,
	REG_HREF,0xA4, // was 0x24 previously, now a HREF offset of 2 was introduced
	REG_VSTART,0x02,
	REG_VSTOP,0x7a,
	REG_VREF,0x0a,
	0x70,0x20,
	0x71,0x20,
	
	0xA2, 0x02, //PCLK delay

	//	from Linux driver
	// Gamma curve values

	0x7a, 0x20 ,
	0x7b, 0x10 ,
	0x7c, 0x1e ,
	0x7d, 0x35 ,
	0x7e, 0x5a ,
	0x7f, 0x69 ,
	0x80, 0x76 ,
	0x81, 0x80 ,
	0x82, 0x88 ,
	0x83, 0x8f ,
	0x84, 0x96 ,
	0x85, 0xa3 ,
	0x86, 0xaf ,
	0x87, 0xc4 ,
	0x88, 0xd7 ,
	0x89, 0xe8 ,

	// AGC and AEC parameters.  Note we start by disabling those features,
	// then turn them only after tweaking the values.
	REG_COM8, COM8_FASTAEC | COM8_AECSTEP ,
	REG_GAIN, 0 ,
	REG_AECH, 0 ,
	REG_COM4, 0x40 , // magic reserved bit
	REG_COM9, 0x18 , // 4x gain + magic rsvd bit
	REG_BD50MAX, 0x05 ,
	REG_BD60MAX, 0x07 ,
	REG_AEW, 0x95 ,
	REG_AEB, 0x33 ,
	REG_VPT, 0xe3 ,
	REG_HAECC1, 0x78 ,
	REG_HAECC2, 0x68 ,
	0xa1, 0x03 , // magic
	REG_HAECC3, 0xd8 ,
	REG_HAECC4, 0xd8 ,
	REG_HAECC5, 0xf0 ,
	REG_HAECC6, 0x90 ,
	REG_HAECC7, 0x94 ,
	REG_COM8, COM8_FASTAEC|COM8_AECSTEP|COM8_AGC|COM8_AEC ,
	0x30,0,
	0x31,0,//disable some delays

	// Almost all of these are magic "reserved" values.
	REG_COM5, 0x61 ,
	REG_COM6, 0x4b ,
	0x16, 0x02 ,
	REG_MVFP, 0x07 ,
	0x21, 0x02 ,
	0x22, 0x91 ,
	0x29, 0x07 ,
	0x33, 0x0b ,
	0x35, 0x0b ,
	0x37, 0x1d ,
	0x38, 0x71 ,
	0x39, 0x2a ,
	REG_COM12, 0x78 ,
	0x4d, 0x40 ,
	0x4e, 0x20 ,
	REG_GFIX, 0 ,
	//0x6b, 0x4a ,
	0x74,0x10,
	0x8d, 0x4f ,
	0x8e, 0 ,
	0x8f, 0 ,
	0x90, 0 ,
	0x91, 0 ,
	0x96, 0 ,
	0x9a, 0 ,
	0xb0, 0x84 ,
	0xb1, 0x0c ,
	0xb2, 0x0e ,
	0xb3, 0x82 ,
	0xb8, 0x0a ,

	// More reserved magic, some of which tweaks white balance
	0x43, 0x0a ,
	0x44, 0xf0 ,
	0x45, 0x34 ,
	0x46, 0x58 ,
	0x47, 0x28 ,
	0x48, 0x3a ,
	0x59, 0x88 ,
	0x5a, 0x88 ,
	0x5b, 0x44 ,
	0x5c, 0x67 ,
	0x5d, 0x49 ,
	0x5e, 0x0e ,
	0x6c, 0x0a ,
	0x6d, 0x55 ,
	0x6e, 0x11 ,
	0x6f, 0x9e , // it was 0x9F "9e for advance AWB"
	0x6a, 0x40 ,
	REG_BLUE, 0x40 ,
	REG_RED, 0x60 ,
	REG_COM8, COM8_FASTAEC|COM8_AECSTEP|COM8_AGC|COM8_AEC|COM8_AWB ,

	// Matrix coefficients
	0x4f, 0x80 ,
	0x50, 0x80 ,
	0x51, 0 ,
	0x52, 0x22 ,
	0x53, 0x5e ,
	0x54, 0x80 ,
	0x58, 0x9e ,

	REG_COM16, COM16_AWBGAIN ,
	REG_EDGE, 0 ,
	0x75, 0x05 ,
	REG_REG76, 0xe1 ,
	0x4c, 0 ,
	0x77, 0x01 ,
	REG_COM13, //0xc3
	0x48 ,	0x4b, 0x09 ,
	0xc9, 0x60 ,		//REG_COM16, 0x38 ,
	0x56, 0x40 ,

	0x34, 0x11 ,
	REG_COM11, COM11_EXP|COM11_HZAUTO ,
	0xa4, 0x82, //Was 0x88
	0x96, 0x00 ,
	0x97, 0x30 ,
	0x98, 0x20 ,
	0x99, 0x30 ,
	0x9a, 0x84 ,
	0x9b, 0x29 ,
	0x9c, 0x03 ,
	0x9d, 0x4c ,
	0x9e, 0x3f ,
	0x78, 0x04 ,

	/* Extra-weird stuff.  Some sort of multiplexor register */
	0x79, 0x01 ,
	0xc8, 0xf0 ,
	0x79, 0x0f ,
	0xc8, 0x00 ,
	0x79, 0x10 ,
	0xc8, 0x7e ,
	0x79, 0x0a ,
	0xc8, 0x80 ,
	0x79, 0x0b ,
	0xc8, 0x01 ,
	0x79, 0x0c ,
	0xc8, 0x0f ,
	0x79, 0x0d ,
	0xc8, 0x20 ,
	0x79, 0x09 ,
	0xc8, 0x80 ,
	0x79, 0x02 ,
	0xc8, 0xc0 ,
	0x79, 0x03 ,
	0xc8, 0x40 ,
	0x79, 0x05 ,
	0xc8, 0x30 ,
	0x79, 0x26 ,

	REG_COM1, 0x0 ,
	REG_COM15, COM15_RGB565|COM15_R00FF,
	REG_COM9, 0x6A ,	 // 128x gain ceiling; 0x8 is reserved bit
	0x4f, 0xb3 ,		 // "matrix coefficient 1"
	0x50, 0xb3 ,		 // "matrix coefficient 2"
	0x51, 0	,		 // vb
	0x52, 0x3d ,		 // "matrix coefficient 4"
	0x53, 0xa7 ,		 // "matrix coefficient 5"
	0x54, 0xe4 ,		 // "matrix coefficient 6"
	REG_COM13, COM13_UVSAT, // |COM13_GAMMA
	0xFF, 0xFF //end marker
};

volatile char image_state = 0; // controls what needs to be done in interrupt routine INT2


// ********************************************************************************************
// ISR (INT2_vect)
// is executed on positive slope of Vsync frame signal of OV7670 camera
// by hardwiring the write reset of the FIFO is activated
// by software we activate the FIFO write enable here
// ********************************************************************************************
ISR (INT2_vect)
{
	if (image_state == 0) {
//		start to store image to FIFO, write enable
		PORTD |= _BV(PD3); 
//		at next VSync stop image saving
		image_state++; }
	else if (image_state == 1) {
//		stop to store image, write disable
		PORTD &= ~(_BV(PD3)); 
//		image is ready for reading from FIFO
		image_state++; }
}

static void
CAM_write_Reg(uint8_t reg_addr, uint8_t reg_val)
{
	i2c2_start(CAM_I2C_Device+I2C_WRITE);
	i2c2_write(reg_addr); // access register
	i2c2_write(reg_val); // write config value
	i2c2_stop();
}

static uint8_t
CAM_read_Reg(uint8_t reg_addr)
{
	uint8_t reg_val=0;
	
	i2c2_start(CAM_I2C_Device+I2C_WRITE);
	i2c2_write(reg_addr); // access register
	i2c2_stop();

	i2c2_start(CAM_I2C_Device+I2C_READ);
	reg_val = i2c2_readNak();
	i2c2_stop();
	return reg_val;
}


void
CAM_init(void)
{
	uint8_t reg_val = 0;	// register value read from CAM
	uint8_t reg_add = 0;	// register address
	const unsigned char *flash_pointer = cam_init_data;
	
	//	initialize software sccb bus for camera register access
	//	with hard wired 5k pull-up-resistors to 3.3V
	i2c2_init();

	//	100s after power-on both FIFO pointers must be reset
	//	i.e. set low and at least two clock periods must occur
	//	for the FIFO write reset this is done automatically
	//	for the FIFO read reset RCLK must be toggled
	//	FIFO_read_reset();
	do
	{
		FIFO_read_reset();

		//	init of camera

		CAM_write_Reg(REG_COM7, COM7_RESET); // reset CAM registers
		lcd_wait_ms(100);

		while (reg_add < 0xFF)
		{
			reg_add = pgm_read_byte(flash_pointer++);
			reg_val = pgm_read_byte(flash_pointer++);
			CAM_write_Reg(reg_add, reg_val);
		}
		
		reg_val = CAM_read_Reg(REG_RGB444);
	}	while (reg_val==0xFF);
}

void
FIFO_read_reset(void)
{
	uint8_t i=0;
	
	PORTF &= ~(_BV(PF2)); //FIFO read reset low
	
	do
	{
		PORTF &= ~(_BV(PF4)); // FIFO Read Clock low
		PORTF |= _BV(PF4); //FIFO Read Clock high
		i++;
	}	while ( i < 2 );

	PORTF &= ~(_BV(PF4)); // FIFO Read Clock low
	PORTF |= _BV(PF2); // FIFO read reset high
	PORTF |= _BV(PF4); // FIFO Read Clock high
}

void capture_frame (const uint8_t pixel_format)
{
	PORTF |= _BV(PF3); //FIFO output disable

	if (pixel_format == yuv422)
	{
//		Change the camera output format to YUV instead of RGB to make brightness value available (Y is always the first of two pixel bytes)
//		The second pixel byte (alternating U and V) can be discarded (color information)
//		COM7 and COM15 must be changed

		CAM_write_Reg(REG_COM7, 0x00); // select YUV output for Camera
		CAM_write_Reg(REG_COM15, 0x00); // deselect RGB565 output format and full output range
	}

	else if (pixel_format == rgb565)
	{
		CAM_write_Reg(REG_COM7, 0x04); // select RGB output for Camera
		CAM_write_Reg(REG_COM15, 0xD0); // select RGB565 output format and full output range
	}

	EIFR |= _BV(INTF2); //clear interrupt flag
	//		activate interrupt INT2 (on PD2)
	EIMSK = _BV(INT2);
	//		now VSYNC rising edge generates an interrupt which causes a FIFO write enable
	image_state = 0; //store new image

	while (image_state != 2) sleep_mode(); //wait until video frame is stored in FIFO

	//		Now one video frame is available in FIFO
	//		now FIFO writing is disabled and we can read out the FIFO
	//		deactivate interrupt INT2 (on PD2)
	EIMSK &= ~(_BV(INT2));

//	Reset FIFO to read frame
	PORTF &= ~(_BV(PF3)); //FIFO output enable
	FIFO_read_reset();
}

