/*****************************************************************************
 
 i2clcd.c - LCD over I2C library 
                Designed for HD44870 based LCDs with I2C expander PCF8574X
                on Atmels AVR MCUs
 
 Copyright (C) 2006 Nico Eichelmann and Thomas Eichelmann
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
 You can contact the authors at info@computerheld.de
 
*****************************************************************************/

/*
        Version 0.4 adapted by Dirk Uffmann on November 26th, 2013
		
		In this version of the library the calling program has to
		issue I2C start and stop conditions:
		
		With lcd_init the first start condition is given. If another I2C 
		device shall be addressed, first an I2C stop condition has to be
		issued. Then an new I2C start has to be issued for the lcd and so on...
		
        Requires I2C-Library from Peter Fleury http://jump.to/fleury
        
        See i2clcd.h for description and example.
*/

#include "i2clcd.h"

char control = LCD_BL; //global variable for lcd control bits
static const char hex[] PROGMEM="0x";
static const char floatspace[] PROGMEM=" ";

//-     Display initialization sequence
void lcd_init(void)
{
		i2c_start_wait(LCD_I2C_DEVICE+I2C_WRITE);
		lcd_write(0x00);				//-		toggle Enable high -> low
        lcd_wait_ms(16);                  //-     Wait for more than 15ms after VDD rises to 4.5V
        lcd_write(LCD_D5 | LCD_D4);     //-     Set interface to 8-bit
        lcd_wait_ms(5);                 //-     Wait for more than 4.1ms
        lcd_write(LCD_D5 | LCD_D4);     //-     Set interface to 8-bit
        lcd_wait_ms(2);               //-     Wait for more than 100us        
        lcd_write(LCD_D5 | LCD_D4);     //-     Set interface to 8-bit
//		while (lcd_busy());				//-		polling of busy flag
		
        lcd_write(LCD_D5);              //-     Set interface to 4-bit
        //- From now on in 4-bit-Mode

        lcd_command(LCD_4BIT | LCD_2LINE | LCD_5X7);            //-     2-Lines, 5x7-Matrix
		lcd_command(LCD_DISPLAYOFF);                            //-     Display off
		lcd_command(LCD_CLEAR);                                 //-     Clear Screen
		lcd_wait_ms(3);											//-		the clear screen command takes 1,5 ms
		lcd_command(LCD_INCREASE | LCD_DISPLAYSHIFTOFF);        //-     Entrymode (Display Shift: off, Increment Address Counter)
		lcd_command(LCD_DISPLAYON);                             //-     Display on
}

//-     Write byte to display with toggle of enable-bit
void lcd_write(char value)
{
	value |= control; // set control bits for backlight control and / or RS
	i2c_write(value | LCD_E);	//- Send data, set enable to high);
	i2c_write(value);			//- toggle enable to low
}

//-		split one byte to two bytes containing the 4-bit data nibbles 
//		and write to lcd
void lcd_split_byte(char value)
{
	char lcddata;
	// used to store value mapped to two bytes
	// lcddata contains one nibble (high or low) of value in high nibble
	// the low nibble of lcddata is used for control bits (RS)

	lcddata = value;
	lcddata &= 0xF0; // mask high nibble
	lcd_write(lcddata);	//- Send data
	lcddata = value;
	lcddata <<= 4; // shift lower nibble of value
	lcddata &= 0xF0; // mask high nibble
	lcd_write(lcddata);	//- Send data
}

//-     Issue a command to the display (use the defined commands from header file)
void lcd_command(char command)
{
	//	while (lcd_busy());				// polling of busy flag

	control = LCD_BL;
	lcd_split_byte(command);	//- Send data
}

//-     Put char to cursor position
void lcd_putchar(char value)
{
	control = LCD_BL | LCD_RS;
	lcd_split_byte(value);	//- Send data
}

//-     Print string to cursor position
void lcd_print( char *string)
{
	unsigned char i = 0;
	while(string[i] != 0x00)
	{
		lcd_putchar(string[i]);
		i++;
	}
}

/*************************************************************************
Display string from program memory without auto linefeed
Input:     string from program memory be be displayed
Returns:   none
*************************************************************************/
void lcd_print_p(const char *progmem_s)
/* print string from program memory on lcd (no auto linefeed) */
{
	register char c;

	while ( (c = pgm_read_byte(progmem_s++)) ) 
	{
		lcd_putchar(c);
	}

} /* lcd_print_p */


//-     Put char to position
bool lcd_putcharlr( unsigned char line,  unsigned char row,  char value)
{
        if(!lcd_gotolr(line, row)) return false;
        lcd_putchar(value);
        
        return true;
}



//-     Print string to position (If string is longer than LCD_ROWS overwrite first chars)(line, row, string)
bool lcd_printlc( unsigned char line, unsigned char row,  char *string)
{
        unsigned char i;
                
        for(i = 0; string[i] != 0x00; i++)
        {
                if (!lcd_putcharlr(line, row, string[i])) return false;
                row++;
                if(row > LCD_ROWS)
                {
                        row = 1;
                }
        }
        return true;
}

//-     Print string to position (If string is longer than LCD_ROWS continue in next line)(line, row, string)
bool lcd_printlrc(unsigned char line, unsigned char row,  char *string)
{
        unsigned char i;
                
        for(i = 0; string[i] != 0x00; i++)
        {
                if(!lcd_putcharlr(line, row, string[i])) return false;
                row++;
                if(row > LCD_ROWS)
                {
                        line++;
                        row = 1;
                }
                if(line > LCD_LINES)
                {
                        line = 1;
                }
        }
        return true;
}

//-     Print string to position (line, row, string)
bool lcd_printlr( unsigned char line,  unsigned char row,  char *string)
{
        if(!lcd_gotolr(line, row)) return false;
        lcd_print(string);
        
        return true;
}

//-     Go to position (line, row)
bool lcd_gotolr( unsigned char line,  unsigned char row)
{
        char lcddata;
        
        if(line > LCD_LINES) return false;
        if(row > LCD_ROWS) return false;
        if((line == 0) || (row == 0) ) return false;
                
        lcddata = LCD_D7;
#if LCD_LINES>=2
        if (line == 2) lcddata = LCD_D7 |  (LCD_LINE2);
#endif
#if LCD_LINES>=3
        if (line == 3) lcddata = LCD_D7 |  (LCD_LINE3);
#endif
#if LCD_LINES>=4
        if (line == 4) lcddata = LCD_D7 |  (LCD_LINE4);
#endif
        lcddata += (row-1);

        lcd_command(lcddata);

        return true;    
}

//-     Go to nextline (if next line > LCD_LINES return false)
bool lcd_nextline(void)
{
        unsigned char line = 0x00, row = 0x00;
        lcd_getlr(&line, &row);
        if (!lcd_gotolr(line + 1, 1)) return false;
                else return true;
}

//-     Read data from i2c
unsigned char lcd_read_i2c(void)
{
        unsigned char lcddata = 0x00;
        i2c_start_wait(LCD_I2C_DEVICE+I2C_READ);
        lcddata = i2c_readNak();
        i2c_stop();
        return lcddata;
}

//-     Write data to i2c
void lcd_write_i2c(unsigned char value)
{
	value |= control; // set control bits for backlight control and / or RS
	i2c_start_wait(LCD_I2C_DEVICE+I2C_WRITE);
	i2c_write(value);
	i2c_stop();
}

//-     Read data from display over i2c (second nibble corresponds to highbyte or lowbyte)
unsigned char lcd_read(bool mode)
{
        unsigned char lcddata;
        if(mode == LCD_DATA)
                lcd_write_i2c(LCD_E | LCD_RS | LCD_RW | LCD_D4 | LCD_D5 | LCD_D6 | LCD_D7);
        else
                lcd_write_i2c(LCD_E | LCD_RW | LCD_D4 | LCD_D5 | LCD_D6 | LCD_D7);
        
        lcddata = lcd_read_i2c();
        lcd_write_i2c(0x00);
        return lcddata;
}

//-     Read one complete byte over i2c from display
unsigned char lcd_getbyte(bool mode)
{
        unsigned char hib, lob;
        hib = lcd_read(mode);
        lob = lcd_read(mode);
        return (hib << 4) + (lob & 0x0F);
}

//-     Get line and row (target byte for line, target byte for row)
bool lcd_getlr(unsigned char *line, unsigned char *row)
{
        unsigned char lcddata, tmp;
        lcddata =       lcd_getbyte(LCD_ADDRESS);
        if (lcddata & (1 << 7)) return false;
        tmp = lcddata;
        tmp &= 0xF0;
        switch(tmp)
        {
                case LCD_LINE1:
                        *line = 1;
                        break;
#if LCD_LINES>=2
                case LCD_LINE2:
                        *line = 2;
                        break;
#endif
#if LCD_LINES>=3
                case LCD_LINE3:
                        *line = 3;
                        break;
#endif
#if LCD_LINES>=4
                case LCD_LINE4:
                        *line = 4;
                        break;
#endif
                default:
                        return false;
        }
        lcddata &= 0x0F;
        *row = lcddata + 1;
        return true;
}

//-     Check if busy
bool lcd_busy(void)
{
        unsigned char state = lcd_getbyte(LCD_ADDRESS);
        if (state & (1 << 7)) return true;
                else return false;
}

/*-     Wait some microseconds
void lcd_wait_us(unsigned short us)
{
        unsigned short i;
        for(i = 0; i < us; i++)
        {
                wait1us;
        }
}
*/
//-     Wait multiples of 1 ms
void lcd_wait_ms( unsigned short ms)
{
	uint8_t counter;
	for (counter=0; counter < ms; counter++)
	{
		timer0=0; // timer0 is incremented every 111s
		while (timer0 < 9) sleep_mode(); // 9 repetitions make 1ms 		
	}
}

void LC_display_dec (uint16_t value, uint8_t line, uint8_t row)
{
	char buffer[6];
	utoa(value, buffer, 10); //display decimal value
	// put string to display
	lcd_gotolr(line,row);
	//	lcd_print_p(hex);
	lcd_print(buffer);
	lcd_print_p(floatspace);
}

void LC_display_hex (uint8_t value, uint8_t line, uint8_t row)
{
	char buffer[6];
	utoa(value, buffer, 16); //display hex value
	// put string to display
	lcd_gotolr(line,row);
//	lcd_print_p(hex);
	lcd_print(buffer);
	lcd_print_p(floatspace);
}

void LC_display_hex_n (uint8_t * start_address, uint8_t number_of_bytes, uint8_t line, uint8_t row)
{
	char buffer[6];

	lcd_gotolr(line,row);

	do {
		utoa(*start_address++, buffer, 16); //display hex value
		// put string to display
		lcd_print(buffer);
		lcd_print_p(floatspace);
		number_of_bytes--;
	} while (number_of_bytes);
}
