//************************************************************************************
//* C-code source file "image_processing.c" for identification of 
//* two figures or ciphers in an image taken by a camera
//*
//* Created: 17.09.2014 16:56:55
//* Last modification: 11.12.2014
//*
//* Author: Dirk Uffmann (uffmann@arcor.de)
//*
//* This program should run on an Arduino Mega 2560 R3 board with an Lseeduino TFT
//* shield V1.3 and 2.4 inch TFT V2.1 with 320x240 pixel.
//* Intention is to test an OV7670 Camera module by Alientek V2.2 with FIFO AL422B.
//* For debugging purpose a text LCD with I2C bus adapter is connected via TWI.
//*
//* The hardware and software is documented in an article provided in the internet
//* magazine ft:pedia 04/2014 by www.ftcommunity.de in german language
//*
//* Title of the article: "Einfache Ziffernerkennung mit einer CMOS-Kamera am AVR-
//* Controller am Beispiel eines ft-Kreditkartenleser-Modells"
//*
//* This software is provided as is without any warranty, guarantee or liability.
//* The author takes no responsibility for consequences or damage caused by the usage
//* of this software.
//*
//************************************************************************************

#include "image_processing.h"

const uint8_t cypher[10][4] PROGMEM = {
	//	left, right, top,  bottom
	{0x17, 0x14, 0x0B, 0x09},	//cypher 0
	{0x03, 0x16, 0x06, 0x01},	//cypher 1
	{0x0d, 0x0e, 0x0e, 0x07},	//cypher 2
	{0x0c, 0x12, 0x0f, 0x0d},	//cypher 3
	{0x0a, 0x14, 0x06, 0x0e},	//cypher 4
	{0x14, 0x0b, 0x13, 0x0c},	//cypher 5
	{0x17, 0x0e, 0x13, 0x0e},	//cypher 6
	{0x0a, 0x0d, 0x11, 0x04},	//cypher 7
	{0x16, 0x13, 0x10, 0x10},	//cypher 8
	{0x15, 0x12, 0x0b, 0x16}	//cypher 9
};

unsigned int identify_figures (const uint8_t threshold_absolute)
{
	//	stores the brightness value (Y of YUV) of one pixel
	uint8_t pixel_byte = 0;
	uint8_t line = 0; // line counter from 0 to 119
	uint8_t byte = 0; // byte counter from 0 to 159

	//	fields for storage of region label, x and y values of all relevant pixels which are connected
	//	assumption is that only ca. 5% of all pixels are filtered out (25% in each dimension, 5% area covered by filtered pixels)
	 static uint8_t region[256] = { [0 ... 255] = 0x00};
	 static uint8_t x[256] = { [0 ... 255] = 0x00};
	 static uint8_t y[256] = { [0 ... 255] = 0x00};
	 static uint8_t transition_counter_x_left[16]; //counts the transitions in horizontal line from black to white for each region
	 static uint8_t transition_counter_x_right[16]; //counts the transitions in horizontal line from black to white for each region
	 static uint8_t transition_counter_y_top[16]; //counts the transitions in vertical line from black to white for each region
	 static uint8_t transition_counter_y_bottom[16]; //counts the transitions in vertical line from black to white for each region
	 static uint8_t max_x[16]; //width of figure, x-value
	 static uint8_t max_y[16]; //height of figure, y-value
	 static uint8_t center_x[16]; //center of figure, x-value
	 static uint8_t center_y[16]; //center of figure, y-value
	 static uint8_t min_x[16]; //origin of figure, x-value
	 static uint8_t min_y[16]; //origin of figure, y-value

	uint8_t region_max=0;

	//	fields containing connected elements for character recognition
	uint8_t filtered_pixel_counter = 0;
	uint8_t label = 0;
	uint8_t i = 0;
	uint8_t j = 0;
	//	uint8_t k = 0;

	int16_t temp_integer;
	int16_t error [16];
	int16_t min_error;
	uint8_t figure;
	uint8_t identified_figure[16];
	uint16_t result;

	TFT_write_Reg(0x20, 0x00, 0x00); // GRAM horizontal Address
	TFT_write_Reg(0x21, 0x00, 0x00); // GRAM Vertical Address
	TFT_write_index(0x22); // set index to TFT graphics RAM

	capture_frame(rgb565);

//	loop for reading one frame from FIFO and put it directly to TFT display
	line=0;

	do
	{
		byte = 0;
		
//		loop for reading one line from FIFO and put it directly to TFT display
		do
		{
			//				first byte
			PORTG &= ~(_BV(PG2)); // TFT write strobe low
			PORTG |= _BV(PG2); //TFT write strobe high, take over pixel byte to TFT
			PORTF &= ~(_BV(PF4)); // FIFO Read Clock low
			PORTF |= _BV(PF4); //FIFO Read Clock high, drive out new pixel byte

			//				byte counts the complete pixel
			byte++;

			//				second byte
			PORTG &= ~(_BV(PG2)); // TFT write strobe low
			PORTG |= _BV(PG2); //TFT write strobe high, take over pixel byte to TFT
			PORTF &= ~(_BV(PF4)); // FIFO Read Clock low
			PORTF |= _BV(PF4); //FIFO Read Clock high, drive out new pixel byte
		} while(byte<160);  // 2 bytes per pixel, line of 80 pixel
		line++;
	} while (line<120); // 60 lines

	//		One frame is now on TFT display
	
//	stores one line of the image for edge filtering
//	static uint8_t image_line[160] = { [0 ... 159] = 0x00};
//	if ((pixel_byte < image_line[byte-1] - threshold_delta) || (pixel_byte < image_line[byte] - threshold_delta) || (pixel_byte > image_line[byte-1] + threshold_delta) ||  (pixel_byte > image_line[byte] + threshold_delta))

	capture_frame(yuv422);

	//		loop for reading one image frame from FIFO and processing each pixel byte consecutively

	line = 0;

	do
	{
		//			read pixel by pixel from FIFO and FILTER each pixel with threshold filter and store
		//			sufficient filtered pixels in table.

		byte = 0;

		do
		{
			pixel_byte = PINA; //store image byte (brightness)
			PORTF &= ~(_BV(PF4)); // FIFO Read Clock low
			PORTF |= _BV(PF4); //FIFO Read Clock high, drive out new pixel byte
			
			if ((line > 30) && (line < 90) && (byte > 50) && (byte < 110)) // exclude first line and first byte to prevent overflow of uint8_t
			{
				if ((pixel_byte < threshold_absolute))//
				{
//					store filtered pixels in labeled list
					if (filtered_pixel_counter < 250)
					{
						filtered_pixel_counter ++;
						x[filtered_pixel_counter] = byte;
						y[filtered_pixel_counter] = line;
					}
				}
			}

			//					second byte is driven out and ignored (color info U or V)
			PORTF &= ~(_BV(PF4)); // FIFO Read Clock low
			PORTF |= _BV(PF4); //FIFO Read Clock high, drive out new pixel byte
			byte++;
		} while (byte < 160);  // 1 Y-byte per pixel, line of 160 pixel

		line++; //increment line counter

	} while (line < 120);

	PORTF |= _BV(PF3); //FIFO output disable

	DDRA = 0xFF; //activate output for TFT
	line = 0;
	i = 1;

	do
	{
//		Put processed image to TFT display
		byte = 0;

		do
		{
			//				check, if pixel is listed in table
			if (y [i] == line)
			{
				if (x [i] == byte)
				{
//					if pixel is listed in table it is displayed "white"
					pixel_byte = 0xFF;
					i++;
				}
				else pixel_byte = 0;
			}
			else pixel_byte = 0;

			PORTA = pixel_byte; //put image byte to TFT
			PORTG &= ~(_BV(PG2)); // TFT write strobe low
			PORTG |= _BV(PG2); //TFT write strobe high, take over pixel byte to TFT
//			put the same byte again to TFT display for white or black pixel
			PORTG &= ~(_BV(PG2)); // TFT write strobe low
			PORTG |= _BV(PG2); //TFT write strobe high, take over pixel byte to TFT
			byte++;
		} while (byte < 160);

		line++; //increment line counter

	} while (line < 120);

	DDRA = 0; //deactivate output from AVR


//		In a first pass of the pixel table we assign neighbouring pixels in x-direction to the same region label

	i = 2;
	label = 1;
	region [1] = label;

	//		first pass of two passes for neighboring check in x-dimension and preliminary region labeling
	do
	{
		label++;

		if (x[i] == x[i-1] + 1)
		{
			if (y[i] == y[i-1])
			{
				label--;
			}
		}

		region[i] = label;
		i++;

	} while (i <= filtered_pixel_counter);

	//		in the second pass we check for neighboring regions in y-dimension and connect neighboring regions (re-label them)

	i = 1;

	do
	{
		j=i+1;

		do
		{
			if ((x[j] > x[i] - 5) && (x[j] < x[i] + 5 ))
			{
				region[j] = region[i];
			}
			j++;
		} while (j<=filtered_pixel_counter); //
		i++;
	} while (i<filtered_pixel_counter);

//		determine the origin x_min and y_min of each figure (=top and left corner)
//		determine the opposite frame corner x_max and y_max of each figure (=bottom and right corner)

	i = filtered_pixel_counter;

	do
	{
		if (region[i] > region_max) region_max = region[i];
		i--;
	} while (i);

	j = region_max; // start with highest region value

	do
	{
		min_x[j] = 0xFF;
		min_y[j] = 0xFF;
		max_x[j] = 0x00;
		max_y[j] = 0x00;
		j--;
	} while (j);

	i = filtered_pixel_counter;

	do
	{
		j = region[i];
		if (min_x[j] > x[i]) min_x[j] = x[i];
		if (min_y[j] > y[i]) min_y[j] = y[i];
		if (max_x[j] < x[i]) max_x[j] = x[i];
		if (max_y[j] < y[i]) max_y[j] = y[i];
		i--;
	} while (i);

	//		determine the geometric center of each figures frame

	j = region_max;

	do
	{
		center_x[j] = (max_x[j] + min_x[j])/2;
		center_y[j] = (max_y[j] + min_y[j])/2;
		j--;
	} while (j);

	//		The characteristics of the 10 different figures are now determined by counting the transitions between black to white in all lines and in all columns

	i=15;
	do
	{
		transition_counter_x_left[i] = 0;
		transition_counter_x_right[i] = 0;
		transition_counter_y_top[i] = 0;
		transition_counter_y_bottom[i] = 0;
		i--;
	} while (i);

	i=filtered_pixel_counter;
	do
	{
		j = region [i];
	
		if (x[i] < center_x[j])
		{
			transition_counter_x_left[j]++;
			if (x[i]==x[i-1]+1)
			{
				if (y[i]==y[i-1])
				{
					transition_counter_x_left[j]--;
				}
			}
		}
		else
		{
			transition_counter_x_right[j]++;
			if (x[i]==x[i-1]+1)
			{
				if (y[i]==y[i-1])
				{
					transition_counter_x_right[j]--;
				}
			}
		}
		i--;
	} while (i);

	byte = 159;

	do
	{
		i = filtered_pixel_counter;
		line = y[i];

		do
		{
			if (x[i]==byte)
			{
				j = region [i];
			
				if (y[i] < center_y[j])
				{
					transition_counter_y_top[j]++;
					if (y[i]==line-1) transition_counter_y_top[j]--;
				}

				else
				{
					transition_counter_y_bottom[j]++;
					if (y[i]==line-1) transition_counter_y_bottom[j]--;
				}

				line = y[i];
			}
		
			i--;
		
		} while (i);

		byte--;

	} while (byte);

	j = 1;
	do
	{
		i = j*3 + 4;
		LC_display_hex(transition_counter_x_left[j], 1, i);
		LC_display_hex(transition_counter_x_right[j], 2, i);
		LC_display_hex(transition_counter_y_top[j], 3, i);
		LC_display_hex(transition_counter_y_bottom[j], 4, i);
		j++;
	} while (j<4);

	//	Now we compare the objects to the cyphers stored in PROGMEM and identify the figures
	//	This is done using an error function
	//	if the error is at it's minimum and lower than a predefined value, the cypher is identified

	j=region_max;
//	identified_figure[j+1]=0x0F;//error marker

	do
	{
		min_error = 20;
		figure = 10;
		identified_figure[j]=0x0F;//error marker

		do
		{
			figure--;
			temp_integer = (signed int) (transition_counter_x_left[j]);
			temp_integer -= (signed int) (pgm_read_byte (&cypher[figure][0]));
			error[figure] = abs(temp_integer);
			temp_integer = (signed int) (transition_counter_x_right[j]);
			temp_integer -= (signed int) (pgm_read_byte (&cypher[figure][1]));
			error[figure] += abs(temp_integer);
			temp_integer = (signed int) (transition_counter_y_top[j]);
			temp_integer -= (signed int) (pgm_read_byte (&cypher[figure][2]));
			error[figure] += abs(temp_integer);
			temp_integer = (signed int) (transition_counter_y_bottom[j]);
			temp_integer -= (signed int) (pgm_read_byte (&cypher[figure][3]));			
			error[figure] += abs(temp_integer);

			if (error[figure] < min_error)
			{
				min_error = error[figure];
				identified_figure[j] = figure;
			}

		} while (figure);

		j--;
	} while (j);


//	sort figures according to their x-position

	i = 1;

	do
	{
		j = i + 1;
		do
		{
			if (min_x[i] > min_x[j])
			{
				figure = identified_figure [i];
				identified_figure [i] = identified_figure [j];
				identified_figure [j] = figure;
				label = min_x [i];
				min_x [i] = min_x [j];
				min_x [j] = label;
			}
			j++;
		} while (j <= region_max);
		i++;
	} while (i < region_max);

//	return identified figures in sequence left to right

	result = 0;
	
	j = 1;
	
	while (j <= region_max)
	{
		result += identified_figure[j];
		if (j < region_max) result *= 10;
		j++;
	}

	return result;
}